
/****************************************************************************/
/*                                                                          */
/*  XARXES DE COMPUTADORS I                                                 */
/*                                                                          */
/*  Practica 1: Sniffer                                                     */
/*                                                                          */
/*  Fichero logica.c: Agrupa todas las funciones necesarias para llevar a   */  
/*                    cabo las operaciones de filtrado de direcciones MAC,  */
/*                    filtrado de direcciones IP, filtrado de puertos y     */
/*                    visualizacin de los campos deseados por el usuario.  */
/*                                                                          */
/*  Realizado por: Francisco Jose Aguilar Celdran y Sergio Blanco Cuaresma  */
/*                                                                          */       
/****************************************************************************/

#include <stdio.h>
#include "cola.h"

/* ----------------------------------------------- */
/* OBTENER_FICHERO: abre el fichero especificado.  */
/* ----------------------------------------------- */

FILE *obtener_fichero (char *fichero)
{
   FILE *f = NULL; 
   
   f = fopen(fichero,"r");

   if (f == NULL) {
      printf("No se ha podido acceder al fichero %s\n", fichero);
      exit(-1);
   }
   return f;
}

/* -------------------------------------------------------------------------- */
/* LECTURA_FICHERO_FILTRO: lee los datos del fichero especificado, para       */
/*                         el tipo de filtro MAC, IP o puertos, y los         */
/*                         guarda en una cola que se devuelve como resultado. */
/* -------------------------------------------------------------------------- */

Cola *lectura_fichero_filtro (FILE *fichero,int modo)
{
   Cola *cola;
   unsigned char *MAC;
   unsigned char *IP;
   unsigned int *puerto;
   int ret;

   cola = (Cola *)crear_cola();  
   switch (modo)
   {
      case FILTRO_MAC:               /* Lectura de las direcciones MAC del filtro 1 */
         while (!feof(fichero))
         {     
            MAC = (unsigned char *)malloc(6 * sizeof(unsigned char));   
            ret = fscanf(fichero,"%d:%d:%d:%d:%d:%d",
                         &MAC[0],&MAC[1],&MAC[2],&MAC[3],&MAC[4],&MAC[5]);
            if (ret > 0)
               encolar(cola,MAC);
            else
               free(MAC);
         }
         break;
      case FILTRO_IP:                /* Lectura de las direcciones IP del filtro 2 */
         while (!feof(fichero))
         {        
            IP = (unsigned char *)malloc(4 * sizeof(unsigned char));
            ret = fscanf(fichero,"%d.%d.%d.%d",&IP[0],&IP[1],&IP[2],&IP[3]);
            if (ret > 0)
               encolar(cola,IP);
            else
               free(IP);
         }        
         break;
      case FILTRO_PUERTOS:           /* Lectura de los puertos del filtro 3 */
         while (!feof(fichero))
         {        
            puerto = (unsigned int *)malloc(sizeof(unsigned int));
            ret = fscanf(fichero,"%d",&puerto);
            if (ret > 0)
               encolar(cola,(char *)puerto);
            else
               free(puerto);
         }                 
   }
   fclose(fichero);  
   return cola;   
}

/* ------------------------------------------------------------------------- */
/* OBTENER_FICHERO: abre el fichero y obtiene la cola de datos fruto de la   */
/*                  lectura de sus valores. Retorna NULL en caso de error.   */
/* ------------------------------------------------------------------------- */

Cola *obtener_filtro (int tipo_filtro, char *fichero)
{
   FILE *f;
   
   if ((f = obtener_fichero(fichero)) != NULL)
      return lectura_fichero_filtro(f,tipo_filtro);
   else
      return NULL;
}

/* ------------------------------------------------------------------------------- */
/* LECTURA_FICHERO_CONFIGURACION: lee los datos del fichero concretado, guardando  */
/*                                los campos de cada protocolo en la cola          */ 
/*                                correspondiente de la estructura Datos_Sniffer.  */
/* ------------------------------------------------------------------------------- */

Datos_Sniffer *lectura_fichero_configuracion (FILE *fichero)
{
   Datos_Sniffer *colas_datos;
   int caracter;
   unsigned char *linea;
   unsigned char buffer[80];
   unsigned char *campo;
   int tipo_cola,i;
   char *c;

   colas_datos = (Datos_Sniffer *)malloc(sizeof(Datos_Sniffer));
   colas_datos -> Cola_Red = (Cola *)crear_cola();
   colas_datos -> Cola_IP = (Cola *)crear_cola();
   colas_datos -> Cola_TCP = (Cola *)crear_cola();
   colas_datos -> Cola_UDP = (Cola *)crear_cola();
   colas_datos -> Cola_ICMP = (Cola *)crear_cola();
 
   if (!feof(fichero))
      caracter = fgetc(fichero);
   while (!feof(fichero))
   {
      // Se ignoran todas las lineas que se inician con '-' o ' '
      if ((caracter == '-') || (caracter == ' '))
      { 
         while ( ((caracter == '-') || (caracter == ' ')) && (!feof(fichero)) )
         {
            c = fgets(buffer,80,fichero); //saltar_linea(fichero);
            if (c != NULL)
               caracter = buffer[0];
         }
         continue;
      }
      if ((caracter == '1') || (caracter == '2') || (caracter == '3') || 
          (caracter == '4') || (caracter == '5'))
      {
         tipo_cola = atoi(&caracter);
         c = fgets(buffer,80,fichero);
         if(c != NULL)
            caracter = buffer[0];
         continue;
      }
      if (!feof(fichero))
      {  
         campo = (unsigned char *)malloc((strlen(buffer)-1) *sizeof(unsigned char));
         for (i=0;i<strlen(buffer)-1;i++)
            campo[i] = buffer[i];
         switch (tipo_cola)
         {
            case CAMPOS_RED:  // Cola de campos de la trama de red
               encolar(colas_datos -> Cola_Red,campo);
               break;
            case CAMPOS_IP:   // Cola de campos del protocolo IP
               encolar(colas_datos -> Cola_IP,campo);
               break;
            case CAMPOS_TCP:  // Cola de campos del protocolo TCP
               encolar(colas_datos -> Cola_TCP,campo);
               break;
            case CAMPOS_UDP:  // Cola de campos del protocolo UDP
               encolar(colas_datos -> Cola_UDP,campo);
               break;
            case CAMPOS_ICMP: // Cola de campos del protocolo ICMP
               encolar(colas_datos -> Cola_ICMP,campo);
               break;
         }
         c = fgets(buffer,80,fichero);
         if (c != NULL)
            caracter = buffer[0];
      }
   }
   fclose(fichero);
   return colas_datos;
}

/* ------------------------------------------------------------------------------ */
/* OBTENER_CONFIGURACION: abre el fichero y obtiene la estructura Datos_Sniffer   */
/*                      fruto de la lectura de los campos de cada protocolo del   */
/*                      fichero de configuracion. Retorna NULL en caso de error.  */  
/* ------------------------------------------------------------------------------ */

Datos_Sniffer *obtener_configuracion (char *fichero)
{
   FILE *f;
   
   if ((f = obtener_fichero(fichero)) != NULL)
      return lectura_fichero_configuracion(f);
   else
      return NULL;
}

/* -------------------------------------------------------------------------- */
/* DIRECCIONES_MAC_IGUALES: compara si dos direcciones MAC son iguales o no.  */ 
/* -------------------------------------------------------------------------- */
   
int direcciones_MAC_iguales (char *a, char *b)
{
   if ((a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]) && (a[3] == b[3]) &&
       (a[4] == b[4]) &&(a[5] == b[5]))
      return 1;
   else
      return 0;
}

/* ------------------------------------------------------------------------ */
/* DIRECCIONES_IP_IGUALES: compara si dos direcciones IP son iguales o no.  */ 
/* ------------------------------------------------------------------------ */

int direcciones_IP_iguales (char *a, char *b)
{
   if ((a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]) && (a[3] == b[3]))
      return 1;
   else
      return 0;
}

/* ---------------------------------------------------------------------------- */
/* INDICAR_FILTADO: informa por el flujo indicado de que un paquete ha sido     */ 
/*                  filtrado, a causa de uno de los tres tipos de filtros.      */
/* ---------------------------------------------------------------------------- */

void indicar_filtrado (int tipo_filtro, unsigned char *v1, unsigned char *v2,
                       FILE *flujo)
{

   switch(tipo_filtro)
   {
      case FILTRO_MAC:
         fprintf(flujo,"  PAQUETE FILTRADO\n");
         fprintf(flujo," ( DIRECCION MAC ORIGEN == %d:%d:%d:%d:%d:%d )\n",
                 v1[0],v1[1],v1[2],v1[3],v1[4],v1[5]);
         fprintf(flujo," ( DIRECCION MAC DESTINO == %d:%d:%d:%d:%d:%d )\n",
                 v2[0],v2[1],v2[2],v2[3],v2[4],v2[5]);
         break;
      case FILTRO_IP:
         fprintf(flujo,"  PAQUETE FILTRADO\n");
         fprintf(flujo," ( DIRECCION IP ORIGEN == %d.%d.%d.%d )\n",
                 v1[0],v1[1],v1[2],v1[3]);
         fprintf(flujo," ( DIRECCION IP DESTINO == %d.%d.%d.%d )\n",
                 v2[0],v2[1],v2[2],v2[3]);
         break;
      case FILTRO_PUERTOS:
         fprintf(flujo,"  PAQUETE FILTRADO\n");
         fprintf(flujo,"( PUERTO ORIGEN == %d )\n",(unsigned int *)v1);
         fprintf(flujo,"( PUERTO DESTINO == %d )\n",(unsigned int *)v2);
         break;
   }
   fprintf(flujo,"------------------------------------------------------------\n");

}

/* -------------------------------------------------------------------------- */
/* APLICAR_FILTRO: compara los elementos de la cola recibida con el elemento  */ 
/*                 actual. En caso de encontrarlo, no se aplica el filtro.    */
/* -------------------------------------------------------------------------- */

int aplicar_filtro (Cola *cola, unsigned char *datos_actuales, int tipo_filtro)
{
   unsigned char *datos_filtro;
   int direccion_filtrada = 0;
   int encontrado = 0;

   reiniciar_PDI(cola);
   datos_filtro = (unsigned char *)obtener_elemento_PDI(cola);
   while ((datos_filtro != NULL) && (encontrado == 0))
   {
      switch(tipo_filtro)
      {
         case FILTRO_MAC:
            if (direcciones_MAC_iguales(datos_filtro,datos_actuales)) 
               encontrado = 1;
            break;
         case FILTRO_IP:
            if (direcciones_IP_iguales(datos_filtro,datos_actuales)) 
               encontrado = 1; 
            break;
         case FILTRO_PUERTOS:
            if ((unsigned int *)datos_filtro == (unsigned int *)datos_actuales)
               encontrado = 1;
            break;
      }
      avanzar_PDI(cola);
      datos_filtro = (unsigned char *)obtener_elemento_PDI(cola);
   }
   direccion_filtrada = encontrado;
   return direccion_filtrada;
}

/* ------------------------------------------------------------------------------- */
/* CAMPOS_IGUALES: determina si las dos cadenas de caracteres son identicas o no.  */ 
/* ------------------------------------------------------------------------------- */

int campos_iguales (char *a,char *b)
{
   if (strlen(a) == strlen(b))
   {
      if (strcmp(a,b) == 0)
         return 1;
      else
         return 0;
   }
   else 
     return 0;
}

/* -------------------------------------------------------------------------- */
/* FILTRAR_CAMPOS: aplica el filtrado de campos para la cola especificada de  */ 
/*                 la estructura Datos_Sniffer, no mostrando el campo actual  */
/*                 si se encuentra en dicha cola.                             */
/* -------------------------------------------------------------------------- */

int filtrar_campos (Datos_Sniffer *colas_datos, unsigned char *actual, int tipo_cola)
{
   unsigned char *datos_filtro;
   int campo_filtrado = 0;
   int encontrado = 0;

   switch (tipo_cola)
   {
      case CAMPOS_RED:
         reiniciar_PDI(colas_datos->Cola_Red);
         datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_Red);
         while ((datos_filtro != NULL) && (encontrado == 0))
         {
            if (campos_iguales(datos_filtro,actual))
               encontrado = 1;
            avanzar_PDI(colas_datos->Cola_Red);
            datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_Red);
         }
         break;
      case CAMPOS_IP:
         reiniciar_PDI(colas_datos->Cola_IP);
         datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_IP);
         while ((datos_filtro != NULL) && (encontrado == 0))
         {
            if (campos_iguales(datos_filtro,actual))
               encontrado = 1;
            avanzar_PDI(colas_datos->Cola_IP);
            datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_IP);
         }
         break;
      case CAMPOS_TCP:
         reiniciar_PDI(colas_datos->Cola_TCP);
         datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_TCP);
         while ((datos_filtro != NULL) && (encontrado == 0))
         {
            if (campos_iguales(datos_filtro,actual))
               encontrado = 1;
            avanzar_PDI(colas_datos->Cola_TCP);
            datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_TCP);
         }   
         break;
      case CAMPOS_UDP:
         reiniciar_PDI(colas_datos->Cola_UDP);
         datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_UDP);
         while ((datos_filtro != NULL) && (encontrado == 0))
         {
            if (campos_iguales(datos_filtro,actual))
               encontrado = 1;
            avanzar_PDI(colas_datos->Cola_UDP);
            datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_UDP);
         }    
         break;
      case CAMPOS_ICMP:
         reiniciar_PDI(colas_datos->Cola_ICMP);
         datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_ICMP);
         while ((datos_filtro != NULL) && (encontrado == 0))
         {
            if (campos_iguales(datos_filtro,actual))
               encontrado = 1;
            avanzar_PDI(colas_datos->Cola_ICMP);
            datos_filtro = (unsigned char *)obtener_elemento_PDI(colas_datos->Cola_ICMP);
         }   
         break;
   } 
   campo_filtrado = encontrado;
   return campo_filtrado;
}

/* ------------------------------------------------------------------------- */
/* ASIGNAR_CADENA: asigna una cadena s2 a una cadena s1, copiando a s1 los.  */ 
/*                 caracteres de s2 comprendidos entre los dos indices.      */
/* ------------------------------------------------------------------------- */

void asignar_cadena(char *s1,char *s2,int indice_final,int indice_inicial)
{
   int i = 0;
   int indice = indice_inicial;  
   int diferencia = indice_final - indice_inicial;

   while (diferencia > 0)
   {
      s1[i] = s2[indice];
      diferencia--;
      indice++;
      i++;
   }
}

/* ---------------------------------------------------------------------- */
/* CONVERTIR_IP: transforma la IP como cadena de caracteres al formato    */
/*               en que estan almacenadas las direcciones IP en la cola.  */
/* ---------------------------------------------------------------------- */

unsigned char *convertir_IP(char *cadena_ip)
{
   int j,i = 0;
   int val,inicial = 0;
   unsigned char *ip,*aux;

   ip = (char *)malloc(4*sizeof(char));
   aux = (char *)malloc(3*sizeof(char));
   for (j=0;j<3;j++)
   {
      if (i != 0)
      {
         i++;
         inicial = i;
      }
      while (cadena_ip[i] != '.')
         i++;
      strcpy(aux,"   ");
      asignar_cadena(aux,cadena_ip,i,inicial);
      val = atoi(aux);
      ip[j] = val;
   }
   strcpy(aux,"   ");
   asignar_cadena(aux,cadena_ip,strlen(cadena_ip),i+1);
   val = atoi(aux);
   ip[3] = val;
   free(aux);
   return ip;
}



