
/*****************************************************************************/
/*                                                                           */
/*  XARXES DE COMPUTADORS I                                                  */
/*                                                                           */
/*  Practica 1: Sniffer                                                      */
/*                                                                           */
/*  Fichero sniffer.c: Implementa el todo el proceso de obtencion de los     */
/*                     paquetes, extraccion de los campos de cada protocolo, */
/*                     y filtrado, visualizacion y/o almacenamiento de los   */
/*                     datos de los paquetes recibidos. Es importante        */ 
/*                     conocer los parametros de ejecucion para su correcto  */
/*                     uso. Ejecutar la aplicacion sin parametros para ver   */
/*                     una descripcion de cada parametro del programa.       */
/*                                                                           */
/*  Realizado por: Francisco Jose Aguilar Celdran y Sergio Blanco Cuaresma   */
/*                                                                           */       
/*****************************************************************************/

#include <stdio.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/socket.h>
#include <time.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <unistd.h>
#include <signal.h>
#include "cola.h"

#define TCP 1
#define UDP 2
#define ICMP 3

struct ethhdr *ether_header;  // Trama de red (IEEE 802.3) definida en
                              // /usr/include/linux/if_ether.h
struct iphdr *ip_header;      // Paquete IP definido en /usr/include/linux/ip.h
struct tcphdr *tcp_header;    // Paquete TCP definido en /usr/include/linux/tcp.h
struct udphdr *udp_header;
struct icmphdr *icmp_header;
struct iphdr_options *opciones_ip;
int sockfd = 0;

/* ---------------------------------------------------------------------- */
/* CREAR_SOCKET_PROMISCUO: Devuelve el descriptor de un socket promiscuo  */
/*                         asociado al interfaz pasado como argumento     */
/*                         (e.g. eth0, lo)                                */ 
/* ---------------------------------------------------------------------- */

int crear_socket_promiscuo(char *d)
{   
   int fd;   
   struct ifreq ifr;   
   int s;   
  
   fd=socket(AF_INET, SOCK_PACKET, htons(0x800));   
   if(fd < 0)
   {
      perror("cant get SOCK_PACKET socket");
      exit(0);
   }

   strcpy(ifr.ifr_name, d);
   s=ioctl(fd, SIOCGIFFLAGS, &ifr);

   if(s < 0)
   {
      close(fd);
      perror("cant get flags");
      exit(0);
   }

   ifr.ifr_flags |= IFF_PROMISC;
   s=ioctl(fd, SIOCSIFFLAGS, &ifr);
   if(s < 0) 
      perror("cant set promiscuous mode");

   return fd;
}

/* ---------------------------------------------------------------------- */
/* INETADDR: Convierte numero de 32 bits en direccin IP human-readable.  */
/* ---------------------------------------------------------------------- */

char *inetaddr ( u_int32_t ip )
{
   struct in_addr in;
   in.s_addr = ip;
   return inet_ntoa(in);
}

/* ---------------------------------------------------------- */
/* HWADDR: Convierte numero en direccin MAC human-readable.  */
/* ---------------------------------------------------------- */

char *hwaddr (u_char * s)
{
   static char buf[30];
   sprintf (buf, "%02X:%02X:%02X:%02X:%02X:%02X",s[0],s[1],s[2],s[3],s[4],s[5]);
  return buf;
}

FILE *almacen = NULL;

/* ---------------------------------------------------------------------- */
/* CONTROLADOR: funcion que permite finalizar correctamente la ejecucion  */
/*              al pulsar crtl+c.                                         */
/* ---------------------------------------------------------------------- */

void controlador (int numero)
{
   if (almacen != NULL) {
      fclose(almacen);
   }
   if (sockfd != 0) {
      close(sockfd);
   }

	/* Se pone controlador por defecto para ctrl-c */
	signal (SIGINT, SIG_DFL);
   exit(-1);
}

/* ------------------------------------------------------------------------------- */
/* ABORT: funcion que muestra por pantalla la sintaxis de ejecucion del programa.  */
/* ------------------------------------------------------------------------------- */

void abort(){
   printf("Sintaxis: ./sniffer [parametros]\n");
   printf("Parametros:\n");
   printf(" [ -h --> esta ayuda ]\n");
   printf(" [ -c <fichero> --> Fichero de configuracion inicial ]\n");
   printf(" [ -v --> Visualizar los datos de los paquetes ]\n");
   printf(" [ -a <fichero> --> Almacenar los datos de los paquetes en el fichero indicado ]\n");
   printf(" [ -m <fichero> --> Fichero con listado de MACs a filtrar ]\n");
   printf(" [ -i <fichero> --> Fichero con listado de IPs a filtrar ]\n");
   printf(" [ -p <fichero> --> Fichero con listado de puertos a filtrar ]\n");
   printf(" [ -f --> Visualizar y/o almacenar la referencia de los paquetes filtrados\n");
   printf(" [ -d --> Visualizar y/o almacenar los datos del cuerpo del paquete\n");
   printf(" [ -r --> interfaz de red (e.g. -r eth0) ]\n");
   printf("IMPORTANTE: Es obligatorio indicar -r, e.g. -r eth0\n");
   printf("            Es obligatorio indicar -c, e.g. -c config.txt\n");
   printf("            Es obligatorio o bien -a, o bien -v, o bien -a y -v\n");
   printf("Es necesario determinar que campos de cada protocolo se desean visualizar\n");
   printf("y/o almacenar, mediante el fichero de configuracion inicial.\n\n");
   exit(1);
}

/* ------------------- */
/* Programa principal  */
/* ------------------- */

int main(int argc, char *argv[])
{
   char paquete[65535];
   int num_paquetes = 0;
   int len,i,j = 0;  
   struct ifreq if_data;
   u_int8_t myMAC[6];
   u_int32_t myIP, myNETMASK, myBROADCAST, ip;
   FILE *salida_datos;
   int num_salidas = 0;
   char *origen, *destino;
   unsigned int puerto_origen;
   unsigned int puerto_destino;
   int protocolo = 0;
   int visualizar = 0;
   int almacenar = 0;
   int filtro_MAC_activo = 0;
   int filtro_IP_activo = 0;
   int filtro_puertos_activo = 0;
   int ver_paquetes_filtrados = 0;
   int ver_datos_paquete = 0;
   int len_cabeceras;
   Cola *cola_MAC;
   Cola *cola_IP;
   Cola *cola_puertos;
   Datos_Sniffer *colas_campos;
   time_t hora;
   int linea_ocupada;   
   char *interfaz = NULL;  // Interfaz de red que escucharemos
   char *config = NULL;    // Fichero de configuracion inicial
   char *fichero_salida;   // Fichero donde se guarda la salida en caso de -a


   // Cambio de controlador para ctrl-c y asi finalizar correctamente el programa
   // al abortar el bucle de recepcion de paquetes con ctrl-c

   signal (SIGINT, controlador);

   // Extraccion de los parametros de la ejecucion
   // getopt busca los parametros -m -i -p -v -a -h -c -r -f -d donde -m -i -p -c -a -r tienen
   // 1 argumento el argumento se optiene de la variable externa optarg

   int c;
   while ((c = getopt (argc, argv, "m:i:p:r:c:va:hfd")) != -1) 
   {
      switch (c) 
      {
         case 'm':   // Filtro para direcciones MAC
            filtro_MAC_activo = 1;
            cola_MAC = (Cola *)obtener_filtro(FILTRO_MAC,optarg);
            if (cola_MAC == NULL)
            {
               fprintf(stderr,"Error en el acceso al fichero de filtro MAC (%s)\n",optarg);
               exit(-1);
            }
            else
               break;
         case 'i':   // Filtro para direcciones IP
            filtro_IP_activo = 1;
            cola_IP = (Cola *)obtener_filtro(FILTRO_IP,optarg);
            if (cola_IP == NULL)
            {
               fprintf(stderr,"Error en el acceso al fichero de filtro IP (%s)\n",optarg);
               exit(-1);
            }
            else
               break;
         case 'p':   // Filtro para puertos
            filtro_puertos_activo = 1;
            cola_puertos = (Cola *)obtener_filtro(FILTRO_PUERTOS,optarg);
            if (cola_puertos == NULL)
            {
               fprintf(stderr,"Error en el acceso al fichero de filtro puertos (%s)\n",optarg);
               exit(-1);
            }
            else
               break;
         case 'r':   // Interfaz de red a utilizar
            interfaz = optarg;
            break;
         case 'c':   // Fichero de configuracion inicial
            config = optarg;
            break;
         case 'v':   // Visualizacion de los paquetes
            visualizar = 1;
            num_salidas++;
            salida_datos = stdout;
            break;
         case 'a':   // Almacenamiento de los paquetes
            almacenar = 1;
            num_salidas++;
            fichero_salida = optarg;
            break;
         case 'h':   // Visualizacion de la sintaxis del programa
            abort();
            break;
         case 'f':
            ver_paquetes_filtrados = 1;
            break;
         case 'd':
            ver_datos_paquete = 1;
            break;
         case '?':   // Opcion incorrecta
            if (isprint (optopt))
               fprintf (stderr, "\nOpcion invalida `-%c'.\n", optopt);
            else
               fprintf (stderr, "\nCaracter de opcion desconocido `\\x%x'.\n", optopt);
            exit(-1);
         default:
            abort ();
      }
   }

   if (((visualizar == 0) && (almacenar == 0)) || (config == NULL) || (interfaz == NULL))
      abort();
   else if ((almacenar == 1) || (visualizar == 1))
   {
      colas_campos = (Datos_Sniffer *)obtener_configuracion(config);
      if (colas_campos == NULL)
      {
         fprintf(stderr,"\nError en el acceso al fichero de configuracion (%s)\n",config);
         exit(-1);
      }
      if (almacenar == 1)
      {
         almacen = fopen(fichero_salida,"a");   // Si existe el fichero concatena al final
         if (almacen == NULL)
            almacen = fopen(fichero_salida,"w");  // Si no existe crea fichero nuevo
         salida_datos = almacen;
         hora = time(NULL);
         fprintf(almacen,"\n............................................................\n");
         fprintf(almacen,"Datos guardados en %s",asctime(localtime(&hora)));
         fprintf(almacen,"............................................................\n");
         fflush(almacen);
      }
   }
   
   sockfd = crear_socket_promiscuo(interfaz);

   ////////////////////////////////////////////////////////////
   // ------------- INFORMACIN DE LA INTERFAZ ------------- //
   ////////////////////////////////////////////////////////////
   
   // Obtenemos la MAC de nuestra interfaz
   strcpy (if_data.ifr_name, interfaz);
   if (ioctl (sockfd, SIOCGIFHWADDR, &if_data) < 0) 
   {
      perror ("No se puede obtener la direccion MAC de mi interfaz!\n");
      exit(1);
   }
   memcpy (myMAC, if_data.ifr_hwaddr.sa_data, 6);
   
   // Obtenemos la IP de nuestra interfaz

   if (ioctl (sockfd, SIOCGIFADDR, &if_data) < 0) 
   {
      perror ("No se puede obtener la direccion IP de mi interfaz!\n");
      exit(1);
   }
   memcpy ((void *) &ip, (void *) &if_data.ifr_addr.sa_data + 2, 4);
   myIP = ntohl (ip);
  
   // Obtenemos la Mascara de red de nuestra interfaz

   if (ioctl (sockfd, SIOCGIFNETMASK, &if_data) < 0) 
   {
      perror ("No se puede obtener la direccion de la mascara de red de mi interfaz!\n");
      exit(1);
   }
   memcpy ((void *) &ip, (void *) &if_data.ifr_netmask.sa_data + 2, 4);
   myNETMASK = ntohl (ip);
      
   // Obtenemos la direccin BroadCast de nuestra interfaz

   if (ioctl (sockfd, SIOCGIFBRDADDR, &if_data) < 0) 
   {
      perror ("No se puede obtener la direccion broadcast de mi interfaz!\n");
      exit(1);
   }
   memcpy ((void *) &ip, (void *) &if_data.ifr_broadaddr.sa_data + 2, 4);
   myBROADCAST = ntohl (ip);
   
   // Imprimimos la informacin de la interfaz seleccionada por las salidas
   // especificadas por linea de comandos.
   for (i = 0; i < num_salidas; i++)
   {
      // Ha especificado por parametro -a (archivo) y -v (visualizar)
      if (num_salidas == 2)
      {
         // La primera pasada del bucle lo visualizamos todo por stdout
         if (i == 0)
            salida_datos = stdout;
         else
            // La segunda pasada del bucle lo guardamos en el archivo
            salida_datos = almacen;
      }
      fprintf(salida_datos, "\n------------------------------------------------------------");
      fprintf(salida_datos, "\n**********************");
      fprintf(salida_datos, " S N I F F E R *********************** \n"); 
      fprintf(salida_datos, "------------------------------------------------------------\n");
      fprintf(salida_datos, "> Interfaz: %s\n", interfaz);
      fprintf(salida_datos, "> MAC: %s\n", hwaddr (myMAC));
      fprintf(salida_datos, "> IP: %s\n", inetaddr(ip));
      fprintf(salida_datos, "> Netmask: %s\n", inetaddr(ip));
      fprintf(salida_datos, "> Broadcast: %s\n", inetaddr(ip));
      fprintf(salida_datos, "------------------------------------------------------------\n");
   }
   ////////////////////////////////////////////////////////////
   // ----------- FIN INFORMACIN DE LA INTERFAZ ----------- //
   ////////////////////////////////////////////////////////////
   

   /* ----------- Leemos de forma indeterminada --------------- */
   for (;;) 
   {

      // Leemos paquete como conjunto de bytes
      len = read(sockfd, paquete, sizeof(paquete));
    
      // Mapeamos el conjunto de bytes a una estructura de trama de red

      ether_header = (struct ethhdr *) paquete;
   
      if(filtro_MAC_activo)
      {
         destino = (unsigned char *)ether_header->h_dest;
         origen = (unsigned char *)ether_header->h_source;
         if ((aplicar_filtro (cola_MAC,destino,FILTRO_MAC) == 0) &&
             (aplicar_filtro (cola_MAC,origen,FILTRO_MAC) == 0))
         {
            if (ver_paquetes_filtrados){
               if (almacenar == 1) {
                  indicar_filtrado(FILTRO_MAC,origen,destino,almacen,num_salidas);
               }
               if (visualizar == 1) {
                  indicar_filtrado(FILTRO_MAC,origen,destino,stdout,num_salidas);
               }
            }
            continue;         
         }
      } 

      // Mapeamos el conjunto de bytes a una estructura de paquete IP, para eso
      // saltamos la parte que corresponde a la trama de red

      ip_header = (struct iphdr *) (paquete + sizeof(struct ethhdr));
      if(filtro_IP_activo)
      {
         destino = (unsigned char *)convertir_IP(inetaddr(ip_header->daddr));
         origen = (unsigned char *)convertir_IP(inetaddr(ip_header->saddr));
         if ((aplicar_filtro (cola_IP,destino,FILTRO_IP,almacen,num_salidas) == 0) &&
             (aplicar_filtro (cola_IP,origen,FILTRO_IP,almacen,num_salidas) == 0))
         {
            if (ver_paquetes_filtrados){
               if (almacenar == 1) {
                  indicar_filtrado(FILTRO_IP,origen,destino,almacen,num_salidas);
               }
               if (visualizar == 1) {
                  indicar_filtrado(FILTRO_IP,origen,destino,stdout,num_salidas);
               }
            }
            continue;         
         }
      }
      
      if (ip_header->protocol == 1)               // ICMP
      {
         icmp_header = (struct icmphdr *)(paquete + sizeof(struct ethhdr) + 4*ip_header->ihl);
         protocolo = ICMP;
         len_cabeceras = sizeof(struct ethhdr) + 4*ip_header->ihl + 
                              sizeof(struct icmphdr);
      }
      else if (ip_header->protocol == 6)          // TCP
      {
         // Mapeamos el conjunto de bytes a una estructura de paquete TCP, para eso
         // saltamos la parte que corresponde a la trama de red y el paquete IP 
         // (longitud variable)
         tcp_header = (struct tcphdr *) (paquete + sizeof(struct ethhdr) + 4*ip_header->ihl );
         protocolo = TCP;
         if (filtro_puertos_activo)
         {
            puerto_origen = ntohs(tcp_header->source);
            puerto_destino = ntohs(tcp_header->dest);
            if ((aplicar_filtro(cola_puertos,puerto_origen,FILTRO_PUERTOS) == 0) &&
                (aplicar_filtro(cola_puertos,puerto_destino,FILTRO_PUERTOS) == 0))
            {
               if (ver_paquetes_filtrados){
                  if (almacenar == 1) {
                     indicar_filtrado(FILTRO_PUERTOS,puerto_origen,puerto_destino,
                                      almacen,num_salidas);
                  }
                  if (visualizar == 1) {
                     indicar_filtrado(FILTRO_PUERTOS,puerto_origen,puerto_destino,
                                      stdout,num_salidas);
                  }
               }
               continue;
            }
         }
         len_cabeceras = sizeof(struct ethhdr) + 4*ip_header->ihl + 
                              sizeof(struct tcphdr);
      }
      else if (ip_header->protocol == 17)         // UDP
      {
         udp_header = (struct udphdr *) (paquete + sizeof(struct ethhdr) + 4*ip_header->ihl );
         protocolo = UDP;
         if (filtro_puertos_activo)
         {
            puerto_origen = ntohs(udp_header->source);
            puerto_destino = ntohs(udp_header->dest);
            if ((aplicar_filtro(cola_puertos,puerto_origen,FILTRO_PUERTOS) == 0) &&
                (aplicar_filtro(cola_puertos,puerto_destino,FILTRO_PUERTOS) == 0))
            {
               if (ver_paquetes_filtrados){
                  if (almacenar == 1) {
                     indicar_filtrado(FILTRO_PUERTOS,puerto_origen,puerto_destino,
                                   almacen,num_salidas);
                  }
                  if (visualizar == 1) {
                     indicar_filtrado(FILTRO_PUERTOS,puerto_origen,puerto_destino,
                                   stdout,num_salidas);
                  }
               }
               continue;
            }
         }
         len_cabeceras = sizeof(struct ethhdr) + 4*ip_header->ihl + 
                              sizeof(struct udphdr);
      } else {    // Otro protocolo
         continue;
      }

      //Visualizacion y/o almacenamiento en fichero de los datos del paquete

      for (i = 0; i < num_salidas; i++)
      {
         // Ha especificado por parametro -a (archivo) y -v (visualizar)
         if (num_salidas == 2)
         {
            // La primera pasada del bucle lo visualizamos todo por stdout
            if (i == 0)
               salida_datos = stdout;
            else
               // La segunda pasada del bucle lo guardamos en el archivo
               salida_datos = almacen;
         }
         fprintf(salida_datos,"| >> Tamao total del paquete %i Bytes\n", len);

         // Impresion a pantalla o fichero de los campos de la Trama de Red
         
         linea_ocupada = 0;
         if (!cola_vacia(colas_campos->Cola_Red))
         {
            fprintf(salida_datos,"| :: Ethernet ::\n");
            if (filtrar_campos(colas_campos,"dest",CAMPOS_RED) == 1)
            {
               fprintf(salida_datos, "| Dest %s ", hwaddr(ether_header->h_dest));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"source",CAMPOS_RED) == 1)
            {
               fprintf(salida_datos, "| Src %s ", hwaddr(ether_header->h_source));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"protocol",CAMPOS_RED) == 1)
            {
               fprintf(salida_datos, "| Proto %i ", ether_header->h_proto);
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
         }
         linea_ocupada = 0;

         // Impresion a pantalla o fichero de los campos del protocolo IP

         if (!cola_vacia(colas_campos->Cola_IP))
         {      
            fprintf(salida_datos,"| :: IP ::\n");
            if (filtrar_campos(colas_campos,"version",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Version %i ", ip_header->version);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"hl",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| HL %i ", ip_header->ihl);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"tos",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Tos %i ", ip_header->tos);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"length",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Length %i ", ip_header->tot_len);
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
            if (filtrar_campos(colas_campos,"id",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| ID %i ", ip_header->id);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"offset",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Fragment Offset %i ", ip_header->frag_off);
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
            if (filtrar_campos(colas_campos,"ttl",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| TTL %i ", ip_header->ttl);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"protocol",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Proto %i ", ip_header->protocol);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"checksum",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Checksum %i ", ip_header->check);
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
            if (filtrar_campos(colas_campos,"source",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Src %s ", inetaddr(ip_header->saddr));
               fprintf(salida_datos, "|\n");
            }
            if (filtrar_campos(colas_campos,"dest",CAMPOS_IP) == 1)
            {
               fprintf(salida_datos, "| Dest %s ", inetaddr(ip_header->daddr));
               fprintf(salida_datos, "|\n");
            }
         }  
         linea_ocupada = 0;  

         // Impresion a pantalla o fichero de los campos del protocolo TCP

         if ((protocolo == TCP) && (!cola_vacia(colas_campos->Cola_TCP)))
         {
            fprintf(salida_datos,"| :: TCP ::\n");
            if (filtrar_campos(colas_campos,"source",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Src %i ", ntohs(tcp_header->source));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"dest",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Dest %i ", ntohs(tcp_header->dest));
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
            if (filtrar_campos(colas_campos,"sequence",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Sequence %i ", ntohs(tcp_header->seq));
               fprintf(salida_datos, "|\n");
            }
            if (filtrar_campos(colas_campos,"ack",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Ack Sequence %i ", ntohs(tcp_header->ack_seq));
               fprintf(salida_datos, "|\n");
            }
            if (filtrar_campos(colas_campos,"header",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Header %i", tcp_header->res1);
               fprintf(salida_datos," %i", tcp_header->doff);
               fprintf(salida_datos," %i", tcp_header->fin);
               fprintf(salida_datos," %i", tcp_header->syn);
               fprintf(salida_datos," %i", tcp_header->rst);
               fprintf(salida_datos," %i", tcp_header->psh);
               fprintf(salida_datos," %i", tcp_header->ack);
               fprintf(salida_datos," %i", tcp_header->urg);
               fprintf(salida_datos," %i", tcp_header->ece);
               fprintf(salida_datos," %i ", tcp_header->cwr);
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"window",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Window %i ", ntohs(tcp_header->window));
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
            if (filtrar_campos(colas_campos,"checksum",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Checksum %i ", ntohs(tcp_header->check));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"urgent",CAMPOS_TCP) == 1)
            {
               fprintf(salida_datos,"| Urgent Pointer %i ", ntohs(tcp_header->urg_ptr));
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
         }

         // Impresion a pantalla o fichero de los campos del protocolo UDP

         else if ((protocolo == UDP) && (!cola_vacia(colas_campos->Cola_UDP)))
         {
            fprintf(salida_datos,"| :: UDP ::\n");
            if (filtrar_campos(colas_campos,"source",CAMPOS_UDP) == 1)
            {
               fprintf(salida_datos,"| Src %i ", ntohs(udp_header->source));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"dest",CAMPOS_UDP) == 1)
            {
               fprintf(salida_datos,"| Dest %i ", ntohs(udp_header->dest));
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
            if (filtrar_campos(colas_campos,"length",CAMPOS_UDP) == 1)
            {
               fprintf(salida_datos,"| Length %i ", ntohs(udp_header->len));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"checksum",CAMPOS_UDP) == 1)
            {
               fprintf(salida_datos,"| Checksum %i ", ntohs(udp_header->check));
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
            linea_ocupada = 0;
         }

         // Impresion a pantalla o fichero de los campos del protocolo ICMP

         else if ((protocolo == ICMP) && (!cola_vacia(colas_campos->Cola_ICMP)))
         {
            fprintf(salida_datos,"| :: ICMP ::\n");
            if (filtrar_campos(colas_campos,"type",CAMPOS_ICMP) == 1)
            {
               fprintf(salida_datos,"| Type %i ", ntohs(icmp_header->type));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"code",CAMPOS_ICMP) == 1)
            {
               fprintf(salida_datos,"| Code %i ", ntohs(icmp_header->code));
               linea_ocupada++;
            }
            if (filtrar_campos(colas_campos,"checksum",CAMPOS_ICMP) == 1)
            {
               fprintf(salida_datos,"| Checksum %i ", ntohs(icmp_header->checksum));
               linea_ocupada++;
            }
            if (linea_ocupada)
               fprintf(salida_datos, "|\n");
         }

         if (ver_datos_paquete)
         {
            fprintf(salida_datos,"| CUERPO DEL PAQUETE  ");
            fprintf(salida_datos,"(%d bytes cabeceras,%d bytes totales): \n",
                                  len_cabeceras,len);
            for(j=0;j<(len - len_cabeceras);j++){
               if (isprint(paquete[len_cabeceras + j])) {
                  fprintf(salida_datos,"%c",paquete[len_cabeceras + j]);
               }
            }
            fprintf(salida_datos,"\n");
         }

         fprintf(salida_datos,"------------------------------------------------------------\n");
         fflush(salida_datos);
      }
   }

}

